"""
Sample data samples from CSV file with replacement using multiple processors in truly parallel fashion.
Each processor needs to know
  - overall number samples
  - absolute byte positions of data samples' start / end as array byte_pos = [[start0, end0], [start1, end1],...]
Once I got this array:
- Perform train-test split by randomly sampling `test_frac` indices from overall number of samples xn = len(byte_pos).
  Those samples are held out to form test dataset. The other samples form global train dataset.
- Assume p processors each of which holds sub random forest. Each sub forest should train on `train_frac` * xn/p data
  samples, where `train_frac` = 1 - `test_frac`. For this, each processor should sample those `train_frac` * xn/p data
  samples from global train dataset with replacement, i.e., globally, the random forest is trained on `train_frac` * xn
  samples, however, some of them might occur multiple times while others might not occur at all.
To do so, each processor needs to have `byte_pos` array!
"""
import os.path
import pathlib
from typing import BinaryIO, List, Tuple, Union

from mpi4py import MPI
from mpi4py.util.dtlib import from_numpy_dtype
import numpy as np
from sklearn.model_selection import train_test_split


def _determine_line_starts(
    f: BinaryIO, comm: MPI.Comm, byte_displs: np.ndarray, byte_counts: np.ndarray
) -> Tuple[list, bytes, int]:
    """
    Determine line starts in bytes from CSV file.

    Parameters
    ----------
    f : BinaryIO
        The handle of CSV file to read from
    comm : MPI.Comm
        The communicator to use.
    byte_displs : np.ndarray
        The displacements of byte chunk starts on each rank, length is `comm.size`.
    byte_counts : np.ndarray
        The counts of the byte chunk on each rank, length is `comm.size`.

    Returns
    -------
    list
        The line starts in each byte chunk.
    bytes
        The bytes chunk read in on each rank.
    int
        The number of chars used to indicate the line end in the file.
    """
    rank = comm.rank  # Set up communicator stuff.

    # Read bytes chunk and count linebreaks.
    # \r (Carriage Return) : Move cursor to line start w/o advancing to next line.
    # \n (Line Feed) : Move cursor down to next line w/o returning to line start.
    # \r\n (End Of Line) : Combination of \r and \n.

    lineter_len = 1  # Set number of line termination chars.
    line_starts = []  # Set up list to save line starts.
    f.seek(byte_displs[rank])  # Jump to pre-assigned position in file.
    bytes_chunk = f.read(byte_counts[rank])  # Read number of bytes from starting position.

    for pos, l in enumerate(bytes_chunk):  # Determine line breaks in bytes chunk.
        if chr(l) == "\n":  # Line terminated by '\n' only.
            if not chr(bytes_chunk[pos - 1]) == "\r":  # No \r\n.
                line_starts.append(pos + 1)
        elif chr(l) == "\r":
            if (
                pos + 1 < len(bytes_chunk) and chr(bytes_chunk[pos + 1]) == "\n"
            ):  # Line terminated by '\r\n'.
                line_starts.append(pos + 2)
                lineter_len = 2
            else:  # Line terminated by '\r' only.
                line_starts.append(pos + 1)
    return line_starts, bytes_chunk, lineter_len


def _get_byte_pos_from_line_starts(
    line_starts: np.ndarray, file_size: int, lineter_len: int
) -> np.ndarray:
    """
    Get line starts and counts in byte from line starts to read lines via seek and read.

    Parameters
    ----------
    line_starts : np.ndarray
        The absolute positions of line starts in bytes.
    file_size : int
        The absolute file size in bytes.
    lineter_len : int
        The number of line termination characters.

    Returns
    -------
    np.ndarray
        An array containing vectors of (start, count) for lines in byte.
    """
    lines_byte = []  # List for line starts and counts
    for idx in range(len(line_starts)):  # Loop through all line starts.
        if idx == len(line_starts) - 1:  # Special case for last line.
            temp = [
                line_starts[idx],  # Line start in bytes
                file_size
                - line_starts[idx]
                - lineter_len,  # Bytes count of line length via difference
            ]
        else:  # All other lines
            temp = [
                line_starts[idx],  # Line start in bytes
                line_starts[idx + 1]
                - line_starts[idx]
                - lineter_len,  # Bytes count of line length via difference
            ]
        lines_byte.append(temp)
    return np.array(lines_byte)


def _decode_bytes_array(
    byte_pos: np.ndarray, f: BinaryIO, sep: str = ",", encoding: str = "utf-8"
) -> List:
    """
    Decode lines from byte positions and counts.

    Parameters
    ----------
    byte_pos : np.ndarray
        The vectors of line starts and lengths in bytes.
    f : BinaryIO
        The handle of the CSV file to read from.
    sep : char
        The character used in the file to separate entries.
    encoding : str
        The encoding used to decode entries from bytes.

    Returns
    -------
    list
        Values read from CSV file as numpy array entries in float format
    """
    lines = []  # List for saving decoded lines
    byte_pos = byte_pos[
        byte_pos[:, 0].argsort()
    ]  # Sort line starts of data items to read from file in ascending order.
    for item in byte_pos:
        f.seek(item[0])  # Go to line start.
        line = f.read(item[1]).decode(
            encoding
        )  # Read specified number of bytes and decode.
        if len(line) > 0:
            sep_values = [
                float(val) for val in line.split(sep)
            ]  # Separate values in each line.
            line = np.array(
                sep_values
            )  # Convert list of separated values to numpy array.
        lines.append(line)  # Append numpy data entry to output list.
    return lines


def _split_indices_train_test(
    n_samples: int, train_split: float, seed: int
) -> Tuple[int, int, np.ndarray, np.ndarray]:
    """
    Make index-based train test split with shuffle.

    Parameters
    ----------
    n_samples : int
        The overall number of samples in the dataset.
    train_split : float
        The fraction of the dataset used for training (0 < `train_split` < 1)
    seed : int
        The seed used for the random number generator.

    Returns
    -------
    n_train_samples : int
        The number of train samples.
    n_test_samples : int
        The number of test samples.
    train_indices : np.ndarray
        The indices of the samples in the dataset used for training.
    test_indices : np.ndarray
        The indices of the samples in the dataset used for testing.
    """
    n_train_samples = int(
        train_split * n_samples
    )  # Determine number of train samples from split.
    n_test_samples = n_samples - n_train_samples  # Determine number of test samples.
    rng = np.random.default_rng(
        seed=seed
    )  # Set same seed over all ranks for consistent train-test split.
    all_indices = np.arange(0, n_samples)  # Construct array of all indices.
    rng.shuffle(all_indices)  # Shuffle them.
    train_indices = all_indices[
        :n_train_samples
    ]  # First `n_train_samples` indices form train set.
    test_indices = all_indices[n_train_samples:]  # Remaining ones form test set.
    return n_train_samples, n_test_samples, train_indices, test_indices


def load_data_csv_parallel(
    path_to_data: Union[pathlib.Path, str],
    header_lines: int,
    comm: MPI.Comm = MPI.COMM_WORLD,
    train_split: float = 0.75,
    verbose: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Load data from CSV file in truly parallel fashion.

    Parameters
    ----------
    path_to_data : str | pathlib.Path
        The path to the CSV file.
    header_lines : int
        The number of header lines.
    comm : MPI.Comm
        The communicator to use.
    train_split : float
        The train-test split fraction.
    verbose : bool
        The verbosity level.

    Returns
    -------
    train_samples_local : np.ndarray
        The rank-dependent train samples.
    train_targets_local : np.ndarray
        The rank-dependent train targets.
    test_samples : np.ndarray
        The global test samples.
    test_targets : np.ndarray
        The global test targets.
    """
    rank, size = comm.rank, comm.size  # Set up communicator stuff.
    file_size = os.stat(path_to_data).st_size  # Get file size in bytes.

    # Determine displacements + counts of bytes chunk to read on each rank.
    n_bytes_local = file_size // size  # Determine number of bytes to read for each process.
    remainder_bytes = file_size % size  # Determine remainder bytes.
    byte_counts = n_bytes_local * np.ones(  # Construct array with each rank's chunk counts.
        (size,), dtype=int
    )
    if (
        remainder_bytes > 0
    ):  # Equally distribute remainder over respective ranks to balance load.
        byte_counts[:remainder_bytes] += 1
    byte_displs = np.concatenate(  # Determine displs via cumulative sum from counts.
        (np.zeros((1,), dtype=int), np.cumsum(byte_counts, dtype=int)[:-1])
    )

    if rank == 0:
        print(f"File size is {file_size} bytes.")

    if rank == 0 and verbose:
        print(f"Displs {byte_displs}, counts {byte_counts} for reading bytes chunks from file.")

    with open(path_to_data, "rb") as f:  # Open csv file to read from.
        # Determine line starts in bytes chunks on each rank.
        line_starts_local, byte_chunk, lineter_len = _determine_line_starts(f, comm, byte_displs, byte_counts)

        # On rank 0, add very first line.
        if rank == 0:
            line_starts_local = [0] + line_starts_local

        if verbose:
            print(f"[{rank}/{size}]: {len(line_starts_local)} line starts in chunk.")

        # Find correct starting point, considering header lines.
        # All-gather numbers of line starts in each chunk.
        counts_linestarts = np.empty(size, dtype=int)
        comm.Allgather(
            [np.array(len(line_starts_local), dtype=int), MPI.INT], [counts_linestarts, MPI.INT]
        )
        # Determine rank where actual data lines start, i.e., remove ranks only containing header lines.
        cumsum_linestarts = list(np.cumsum(counts_linestarts))
        start_rank = next(i for i in range(size) if cumsum_linestarts[i] > header_lines)
        if verbose:
            print(
                f"[{rank}/{size}]: There are {counts_linestarts} line starts overall.\n"
                f"[{rank}/{size}]: Their cumsum is {cumsum_linestarts}.\n"
                f"[{rank}/{size}]: The start rank is {start_rank}."
            )

        if rank < start_rank:  # Ranks containing only header lines.
            line_starts_local = []
        if rank == start_rank:  # Rank containing header + data lines.
            lines_to_remove = header_lines - (0 if start_rank == 0 else cumsum_linestarts[start_rank - 1])
            line_starts_local = line_starts_local[lines_to_remove:]

        # Share line starts of data samples across all ranks via Allgatherv.
        line_starts_local += byte_displs[
            rank
        ]  # Shift line starts on each rank according to displs.
        if verbose:
            print(
                f"[{rank}/{size}]: {len(line_starts_local)} line starts of shape {line_starts_local.shape} and type "
                f"{line_starts_local.dtype} in local chunk: {line_starts_local}"
            )
        counts_linestarts_corrected = (
            np.empty(  # Initialize array to all-gather local numbers of line starts (corrected for header lines).
                size, dtype=int
            )
        )
        comm.Allgather(
            [np.array(len(line_starts_local), dtype=int), MPI.INT],
            [counts_linestarts_corrected, MPI.INT]
        )
        displs_linestarts_corrected = (
            np.concatenate(  # Determine displacements of line starts from counts.
                (
                    np.zeros((1,), dtype=int),
                    np.cumsum(counts_linestarts_corrected, dtype=int)[:-1],
                ),
                dtype=int,
            )
        )
        if verbose and rank == 0:
            print(
                f"Overall {np.sum(counts_linestarts_corrected)} linestarts.\n"
                f"Number of linestarts in each chunk is {counts_linestarts_corrected}.\n"
                f"Displs of linestarts in each chunk is {displs_linestarts_corrected}."
            )

        line_starts_global = (
            np.empty(  # Initialize array to all-gatherv line starts from all ranks.
                (np.sum(counts_linestarts_corrected),), dtype=line_starts_local.dtype
            )
        )
        if verbose and rank == 0:
            print(
                f"Recvbuf {line_starts_global}, {line_starts_global.shape}, {line_starts_global.dtype}."
            )
        comm.Allgatherv(
            line_starts_local,
            [
                line_starts_global,
                counts_linestarts_corrected,
                displs_linestarts_corrected,
                from_numpy_dtype(line_starts_local.dtype),
            ],
        )
        # Line starts were determined as those positions following a line end.
        # But: There is no line after last line end in file.
        # Thus, remove last entry from all_line_starts.
        line_starts_global = line_starts_global[:-1]
        if rank == 0:
            print(f"After Allgatherv: All line starts: {line_starts_global}")
        # Construct array with line starts and lengths in bytes.
        print(
            f"[{rank}/{size}]: Construct array with line starts and lengths in bytes."
        )
        lines_in_byte = _get_byte_pos_from_line_starts(
            line_starts_global, file_size, lineter_len
        )

        # Make global train-test split.
        print(f"[{rank}/{size}]: Make global train-test split.")
        (
            n_train_samples,
            n_test_samples,
            train_indices,
            test_indices,
        ) = _split_indices_train_test(
            n_samples=len(lines_in_byte), train_split=train_split, seed=size
        )

        n_train_samples_local = (
            n_train_samples // size
        )  # Determine local train dataset size.
        remainder_train = n_train_samples % size  # Balance load.
        if rank < remainder_train:
            n_train_samples_local += 1

        # Construct held-out test dataset (same on each rank).
        print(f"[{rank}/{size}]: Decode {n_test_samples} test samples from file.")
        test_lines = _decode_bytes_array(
            lines_in_byte[test_indices], f, sep=",", encoding="utf-8"
        )
        test_samples = np.array(test_lines)[:, 1:]
        test_targets = np.array(test_lines)[:, 0]
        if verbose:
            print(
                f"[{rank}/{size}]: Test samples: {test_samples[0]}\n"
                f"Test targets: {test_targets[0]}\n"
            )

        # Construct train dataset (different on each rank).
        rng = np.random.default_rng(seed=rank)
        print(f"[{rank}/{size}]: Draw local {n_train_samples_local} train indices.")
        train_indices_local = rng.choice(train_indices, size=n_train_samples_local)
        print(f"[{rank}/{size}]: Decode train lines from file.")
        train_lines_local = _decode_bytes_array(
            lines_in_byte[train_indices_local], f, sep=",", encoding="utf-8"
        )
        train_samples_local = np.array(train_lines_local)[:, 1:]
        train_targets_local = np.array(train_lines_local)[:, 0]

    # Now, each rank holds a local train set (samples + targets)
    # and the global held-out test set (samples + targets).
    return train_samples_local, train_targets_local, test_samples, test_targets


def load_data_csv_root(
    path_to_data: Union[str, pathlib.Path],
    header_lines: int,
    comm: MPI.Comm,
    seed: int,
    train_split: float = 0.75,
    verbose: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Load data from CSV file via root process.

    Parameters
    ----------
    path_to_data : str | pathlib.Path
        The path to the CSV file.
    header_lines : int
        The number of header lines.
    comm : MPI.Comm
        The communicator to use.
    seed : int
        The seed used for train-test split.
    train_split : float
        The train-test split fraction.
    verbose : bool
        The verbosity level

    Returns
    -------
    np.ndarray
        The rank-dependent train samples.
    np.ndarray
        The rank-dependent train targets.
    np.ndarray
        The global test samples.
    np.ndarray
        The global test targets.
    """
    rank, size = comm.rank, comm.size  # Set up communicator stuff.

    if rank == 0:
        data = np.loadtxt(
            path_to_data, dtype=float, delimiter=",", skiprows=header_lines
        )  # Load data into numpy array.
        samples, targets = (
            data[:, 1:],
            data[:, 0],
        )  # Divide data into samples and targets.
        samples_train, samples_test, targets_train, targets_test = train_test_split(
            samples, targets, test_size=1 - train_split, random_state=seed
        )  # Perform train-test split.
        n_train_samples = len(
            samples_train
        )  # Determine number of samples in train set.
        n_test_samples = len(samples_test)  # Determine number of samples in test set.
        n_features = samples_train.shape[1]  # Determine number of features.
        n_train_samples_local = (
            n_train_samples // size
        )  # Determine rank-local number of train samples.
        remainder_train = n_train_samples % size
        train_counts = n_train_samples_local * np.ones(
            size, dtype=int
        )  # Determine load-balanced counts and displacements.
        train_counts[:remainder_train] += 1
        train_displs = np.concatenate(
            (np.zeros(1, dtype=int), np.cumsum(train_counts, dtype=int)[:-1]), dtype=int
        )
        print(
            f"There are {n_train_samples} train and {n_test_samples} test samples.\n"
            f"Local train samples: {train_counts}"
        )
        # Sample `n_train_samples` indices from train set with replacement.
        rng = np.random.default_rng(seed=size)
        train_indices = rng.choice(range(n_train_samples), size=n_train_samples)
        print(f"train_indices have shape {train_indices.shape}.")
        # Construct train dataset from drawn indices (repetitions possible!).
        samples_train_shuffled = samples_train[train_indices]
        targets_train_shuffled = targets_train[train_indices]
        send_buf_train_samples = [
            samples_train_shuffled,
            train_counts * n_features,
            train_displs * n_features,
            from_numpy_dtype(samples_train_shuffled.dtype),
        ]
        send_buf_train_targets = [
            targets_train_shuffled,
            train_counts,
            train_displs,
            from_numpy_dtype(targets_train_shuffled.dtype),
        ]

    else:
        train_counts = np.empty(size, dtype=int)
        n_features = None
        n_test_samples = None
        send_buf_train_samples = None
        send_buf_train_targets = None

    n_features = comm.bcast(n_features, root=0)
    n_test_samples = comm.bcast(n_test_samples, root=0)
    comm.Bcast(train_counts, root=0)
    samples_train_local = np.empty((train_counts[rank], n_features), dtype=float)
    targets_train_local = np.empty((train_counts[rank],), dtype=float)
    recv_buf_train_samples = [
        samples_train_local,
        from_numpy_dtype(samples_train_local.dtype),
    ]
    recv_buf_train_targets = [
        targets_train_local,
        from_numpy_dtype(targets_train_local.dtype),
    ]
    if rank != 0:
        samples_test = np.empty((n_test_samples, n_features), dtype=float)
        targets_test = np.empty((n_test_samples,), dtype=float)
    comm.Scatterv(send_buf_train_samples, recv_buf_train_samples, root=0)
    comm.Scatterv(send_buf_train_targets, recv_buf_train_targets, root=0)
    comm.Bcast(samples_test, root=0)
    comm.Bcast(targets_test, root=0)
    if verbose:
        print(
            f"[{rank}/{size}]: Train samples after Scatterv have shape {samples_train_local.shape}.\n"
            f"Train targets after Scatterv have shape {targets_train_local.shape}.\n"
            f"Test samples after Bcast have shape {samples_test.shape}.\n"
            f"Test targets after Bcast have shape {targets_test.shape}."
        )
    return samples_train_local, targets_train_local, samples_test, targets_test
